/**
 * The <code>AbstractOption</code> class is an abstract
 * class for pricing options.  To use the class, you must
 * derive a class from this class and add code
 * for the following pure virtual functions:
 * <br>
 * <ul>
 * <li> virtual double F(double s, double x);
 * <li> virtual double payoff(double s, double x, int t);
 * </ul> <br>
 * The following member functions will be inherited
 * by the derived class:
 * <br>
 * <ul>
 * <li> void calc_european_option();
 * <li> void genPDAG();
 * <li> void printPDAG();
 * </ul><br>
 * The genPDAG() function should be called before
 * calc_european_option or printPDAG()
 * 
 * Chapter 2 and Chapter 3 of the 46-927 course notes
 * describe pricing an option on a PDAG.  This an 
 * implementation of the process they describe.
 * 
 * @author  Somesh Jha
 * @author  Jeff Stephenson
 */


public abstract class AbstractOption 
{


  /**
   * The (integer) time horizon.
   */
  int T;

  /**
   * The (double) short rate.
   */
  double r;

  /**
   * The (double) up factor.
   */
  double u;
  
  /**
   * The (double) down factor.
   */ 
  double d;

  /**
   * The (double) initial stock price.
   */
  double s_0;

  /**
   * The (double) risk neutral up-tick probability.
   */
  double p;

  /**
   * The (double) risk neutral down-tick probability.
   */	  
  double q;

  /**
   * The array of LinkList objects that represents the 
   * PDAG.  nodes[i] is a linked list of nodes with time
   * value i
   */
  LinkList nodes[];

  /**
   * A boolean flag used for debugging.
   */
  static final boolean DEBUG=false;

  /**
   * A superclass constructor for derived objects, this function
   * sets the initial values of the member fields.
   *
   * @param   T			the (int) time Horizon
   * @param   r			the (double) short rate
   * @param   u			the (double) up-factor
   * @param   d			the (double) down-factor
   * @param   s_0		the (double) initial stock price
   */
  public AbstractOption(int T, double r, double u,
			double d, double s_0) 
  {
    this.T = T;
    this.r = r;
    this.u = u;
    this.d = d;
    this.s_0 = s_0;

    //calculate the risk-nuetral probabilities
    p = (1+r -d)/(u-d);
    q = 1-p;

    //Allocate the link list
    nodes = new LinkList[T+1];
    for(int i=0; i <= T; i++)
    {
        nodes[i] = new LinkList();
    }

  }
    
  /**
   * Determines how the auxillary process evolves
   * over time.  This method is abstract and must
   * be implemented in a derived class.
   *
   * @param   s     the (double) new stock price
   * @param   x	    the (double) old value of the aux process
   */
  public abstract double F(double s, double x);


  /**
   * Determines the payoff in terms of the stock price,
   * auxilliary process and time.   This method is abstract 
   * and must be implemented in a derived class.
   *
   * @param   s     the (double) stock price
   * @param   x	    the (double) value of the aux process
   * @param   t     the (int) time value
   */
  public abstract double payoff(double s,double x, int t);


  /**
   * Generates the PDAG out to the time Horizon T
   * according to the values of the member fields.
   * Must be called before any of the calc methods.
   *
   */
  public void genPDAG() 
  {

    if (DEBUG)
	{
		System.out.println("Entered genPDAG");
	}

    nodes[0].Insert(new Key(0,s_0,s_0));

    for(int t=0; t < T; t++) 
	{
		Node x=nodes[t].head;

		while (x != null) 
		{
	        if (DEBUG) 
            {
	            System.out.println("Considering node: ");
	            x.key.printKey();
	        }
	  
	        double s = ((Key) x.key).stock_price;
	        double a = ((Key) x.key).aux_process;

	        //Generate successor nodes
	        Key up_key = new Key(t+1,s*u,F(s*u,a));
	        Key down_key = new Key(t+1,s*d,F(s*d,a));

	        x.succ[0] = nodes[t+1].Insert(up_key);
	        x.succ[1] = nodes[t+1].Insert(down_key);
	        
	        x = x.next;
        }

        if (DEBUG) 
        {
            printPDAG();
        }
    }
	
    if (DEBUG) 
    {
        System.out.println("Leaving genPDAG");
    }
      
  }

  /**
   * Calculates the option value on the PDAG using the
   * European method where payoff is evaluated only
   * at the time horizon. 
   *
   */
  public void calc_european_option() 
  {

    if (DEBUG) 
    {
        System.out.println("Entered calc_european_option");
    }
    
    //Go backwards in time
    for(int t=T; t >= 0; t--) 
    {
        Node x = nodes[t].head;
        while (x != null) 
        {
            if (t == T)
            {
                ((Key) x.key).option_value = payoff(((Key) x.key).stock_price, ((Key) x.key).aux_process, T);
            }
	        else 
            {
	            //use backward equation
	            double sum = p*((Key) x.succ[0].key).option_value +
	                         q*((Key) x.succ[1].key).option_value;

                ((Key) x.key).option_value = sum/(1+r);
	        }
            x=x.next;
        }
    }

    if (DEBUG)
    {
        System.out.println("Leaving calc_european_option");
    }

  }


  /**
   * Calculates the option value on the PDAG using the
   * American method where payoff can be evaluated at
   * any node
   */
  public void calc_american_option() 
  {
      //Your code here
  }


  /**
   * Prints the PDAG. 
   *
   */
  public void printPDAG() 
  {
      for(int t=0; t <= T; t++) 
      {
          System.out.print("Nodes at time ");
          System.out.println(t);
          System.out.println("------------BEGIN-----------------");

          try
          {
            nodes[t].printList();
          }
          catch (LinkListException e)
          {
            System.out.println("Empty List");
          }

          System.out.println("------------END------------------");
      }
  }

}
  
